package explorviz.hpc_monitoring.reader;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import com.lmax.disruptor.RingBuffer;

import explorviz.hpc_monitoring.disruptor.RecordArrayEvent;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.record.HostApplicationMetaData;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.StringRegistryRecord;
import explorviz.hpc_monitoring.record.TimedPeriodRecord;
import explorviz.hpc_monitoring.record.event.normal.AfterFailedOperationEvent;
import explorviz.hpc_monitoring.record.event.normal.AfterOperationEvent;
import explorviz.hpc_monitoring.record.event.normal.BeforeOperationEvent;

public class TCPReaderThread extends Thread implements IPeriodicTimeSignalReceiver {
	private static final int MESSAGE_BUFFER_SIZE = 131072;
	private static final int OUTPUT_MESSAGE_BUFFER_SIZE = 16384;

	private HostApplicationMetaData hostApplicationMetadata;

	private final static Map<Integer, String> stringRegistry = new TreeMap<Integer, String>();
	private final static List<byte[]> waitingForStringMessages = new ArrayList<byte[]>(1024);

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Received records/sec in Reader" + Thread.currentThread().getId());

	private final SocketChannel socketChannel;

	private final RingBuffer<RecordArrayEvent> ringBuffer;
	private IRecord[] outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE];
	private int outputBufferIndex = 0;

	public TCPReaderThread(final SocketChannel socketChannel,
			final RingBuffer<RecordArrayEvent> ringBuffer) {
		this.socketChannel = socketChannel;
		this.ringBuffer = ringBuffer;

		new TimeSignalReader(1 * 1000, this).start();
	}

	@Override
	public void run() {
		final ByteBuffer buffer = ByteBuffer.allocateDirect(MESSAGE_BUFFER_SIZE);
		try {
			while ((socketChannel.read(buffer)) != -1) {
				buffer.flip();
				messagesfromByteArray(buffer);
			}
		} catch (final IOException ex) {
			System.out.println("Error in read() " + ex.getMessage());
		}
	}

	@Override
	public void periodicTimeSignal(final long timestamp) {
		// TODO flush out buffer!
		final IRecord[] buffer = new IRecord[1];
		buffer[0] = new TimedPeriodRecord();

		final long hiseq = ringBuffer.next();
		final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
		valueEvent.setValues(buffer);
		valueEvent.setValuesLength(1);
		valueEvent.setMetadata(hostApplicationMetadata);
		ringBuffer.publish(hiseq);
	}

	private final void messagesfromByteArray(final ByteBuffer buffer) {
		while (buffer.remaining() > 0) {
			final byte clazzId = buffer.get();
			switch (clazzId) {
				case HostApplicationMetaData.CLAZZ_ID: {
					if (buffer.remaining() >= (HostApplicationMetaData.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInTraceMetadata(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case BeforeOperationEvent.CLAZZ_ID: {
					if (buffer.remaining() >= (BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInBeforeOperationEvent(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case AfterFailedOperationEvent.CLAZZ_ID: {
					if (buffer.remaining() >= (AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInAfterFailedOperationEvent(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case AfterOperationEvent.CLAZZ_ID: {
					if (buffer.remaining() >= (AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInAfterOperationEvent(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case StringRegistryRecord.CLAZZ_ID: {
					int mapId = 0;
					int stringLength = 0;
					if (buffer.remaining() >= 8) {
						mapId = buffer.getInt();
						stringLength = buffer.getInt();
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}

					if (buffer.remaining() >= stringLength) {
						final byte[] stringByteArray = new byte[stringLength];

						buffer.get(stringByteArray);

						addToRegistry(mapId, new String(stringByteArray));
					} else {
						buffer.position(buffer.position() - 9);
						buffer.compact();
						return;
					}
					break;
				}
				default: {
					System.out.println("unknown class id " + clazzId + " at offset "
							+ (buffer.position() - 4));
					buffer.clear();
					return;
				}
			}
		}

		buffer.clear();
	}

	private final void readInTraceMetadata(final ByteBuffer buffer) {
		final int hostnameId = buffer.getInt();
		final int applicationId = buffer.getInt();

		final String hostname = getStringFromRegistry(hostnameId);
		final String application = getStringFromRegistry(applicationId);

		if ((hostname != null) && (application != null)) {
			hostApplicationMetadata = new HostApplicationMetaData(hostname, application);
		} else {
			final byte[] message = new byte[HostApplicationMetaData.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position() - HostApplicationMetaData.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInBeforeOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);

		if (operation != null) {
			putInRingBuffer(new BeforeOperationEvent(timestamp, traceId, orderIndex, operationId,
					operation));
		} else {
			final byte[] message = new byte[BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position() - BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInAfterFailedOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();
		final int causeId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		final String cause = getStringFromRegistry(causeId);

		if ((operation != null) && (cause != null)) {
			putInRingBuffer(new AfterFailedOperationEvent(timestamp, traceId, orderIndex,
					operationId, operation, cause));
		} else {
			final byte[] message = new byte[AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position() - AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInAfterOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		if (operation != null) {
			putInRingBuffer(new AfterOperationEvent(timestamp, traceId, orderIndex, operationId,
					operation));
		} else {
			final byte[] message = new byte[AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position() - AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}

	}

	private final void putInWaitingMessages(final byte[] message) {
		waitingForStringMessages.add(message);
	}

	private final void checkWaitingMessages() {
		final List<byte[]> localWaitingList = new ArrayList<byte[]>();
		for (final byte[] waitingMessage : waitingForStringMessages) {
			localWaitingList.add(waitingMessage);
		}
		waitingForStringMessages.clear();

		for (final byte[] waitingMessage : localWaitingList) {
			final ByteBuffer buffer = ByteBuffer.wrap(waitingMessage);
			final byte waitingMessageClazzId = buffer.get();
			switch (waitingMessageClazzId) {
				case HostApplicationMetaData.CLAZZ_ID:
					readInTraceMetadata(buffer);
					break;
				case BeforeOperationEvent.CLAZZ_ID:
					readInBeforeOperationEvent(buffer);
					break;
				case AfterFailedOperationEvent.CLAZZ_ID:
					readInAfterFailedOperationEvent(buffer);
					break;
				case AfterOperationEvent.CLAZZ_ID:
					readInAfterOperationEvent(buffer);
					break;
				default:
					break;
			}
		}
	}

	private final void putInRingBuffer(final IRecord message) {
		counter.inputObjects(message);

		outputBuffer[outputBufferIndex++] = message;
		if (outputBufferIndex == OUTPUT_MESSAGE_BUFFER_SIZE) {
			flushOutputBuffer();
		}
	}

	private void flushOutputBuffer() {
		if (outputBufferIndex > 0) {
			final long hiseq = ringBuffer.next();
			final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
			valueEvent.setValues(outputBuffer);
			valueEvent.setValuesLength(outputBufferIndex);
			valueEvent.setMetadata(hostApplicationMetadata);
			ringBuffer.publish(hiseq);

			outputBuffer = new IRecord[OUTPUT_MESSAGE_BUFFER_SIZE];
			outputBufferIndex = 0;
		}
	}

	private final void addToRegistry(final int key, final String value) {
		stringRegistry.put(key, value);

		// System.out.println("put key " + key + " value " + value);

		checkWaitingMessages();
	}

	private final String getStringFromRegistry(final int id) {
		return stringRegistry.get(id);
	}
}
